
package info.seasip.jsw64;

import java.io.*;
import java.util.Vector;

public class DragonManic
{
	public static final int AIR     = 0x01; // Blue on black => air
	public static final int EARTH   = 0x0C;	// Green on blue => earth
	public static final int WATER1  = 0x0D;	// Cyan on blue => water
	public static final int WATER2  = 0x4D;	// Bright cyan on blue => water
	public static final int CRUMBLY = 0x0E;	// Yellow on blue => crumbly
	public static final int CONVEYOR= 0x04;	// Green on black => conveyor
	public static final int FIRE1   = 0x02;	// Red on black => fire 1
	public static final int FIRE2   = 0x42;	// Bright red on black => fire 2
	public static final int PORTAL  = 0x4F; // Bright white on blue=> portal
	byte [] dragonMem = new byte[32768];
	int [] spritePage = { 
		0xF6, 0xF5, 0xF4, 0xF3, 0xF2, 0xF1, 0xF0, 0xEF,
		0xEE, 0xED, 0xEA, 0xE9, 0xE8, 0xE7, 0xE6, 0xE5,
		0xE4, 0xE3, 0xE2, 0xE1, 0xE0, 0xDF, 0xDE, 0xDD,
		0xD9, 0xD8, 0xD7, 0xD6, 0xD5, 0xD4, 0xD3, 0xD2,
		0xD1, 0xD0, 0xCF, 0xCE, 0xCD, 0xCC, 0xCB, 0xCA,
		0xC9, 0xC8, 0xC7, 0xC6, 0xC5, 0xBF, 0xBE, 0xBD,
		0xBC, 0xBB, 0xBA, 0xB9, 0xB8, 0xB7, 0xB6, 0xB5,
		0xB4, 0xB3, 0xB2, 0xB1, 0xB0, 0xAF, 0xAE, 0xAD,
		0xAC, 0xAB, 0xA9, 0xA8, 0xA7
	};

	static final byte eugene[] = 
	{
		(byte)0x88, 0x53,(byte)0x85, 0x18, 
		0x02, (byte)0xA2, (byte)0xAE, (byte)0xAE
	};

	static final byte kong[] = 
	{
		(byte)0x98, 0x06, (byte)0xA6, 0x00, 0x01,(byte)0xA6, 0x00, 0x00,
		(byte)0xA8, 0x04, (byte)0x80, 0x10, 0x10,0x00, 0x11, (byte)0xB0,
		(byte)0x98, 0x06, (byte)0xB2, 0x00, 0x01,(byte)0xA6, 0x00, 0x00,
		(byte)0xD8, 0x0C, (byte)0x80, (byte)0x81, 0, 0, 0, 0x12
	};

	// The order of letters in the Dragon font
	static final String charSet = " ABCDEFGHIKLMNOPRSTUVWXYJ\177=0123456789";
	// Sprite frame mapping from the Dragon's 6 Willies to the Spectrum's 8
	private static final int mapFrame[] = { 4, 3, 4, 5, 2, 1, 0, 1 };

	private static final byte typewriter[] = 
	{
		0x0F, 0x1F,     /* T */
		0x1F, 0x0F,     /* Y */
		0x1F, 0x1E,     /* P */
		0x1B, 0x1F,     /* E */
		0x1D, 0x1F,     /* W */
		0x17, 0x1F,     /* R */
		0x1F, 0x1B,     /* I */
		0x0F, 0x1F,     /* T */
		0x1B, 0x1F,     /* E */
		0x17, 0x1F,     /* R */
	};

	Vector spriteMap;
	Jsw64Game target;

	public DragonManic(String s) throws FileNotFoundException, IOException
	{
		target = new Jsw64Game();
		loadSnap(s);
	}

	public DragonManic()
	{
		target = new Jsw64Game();
	}

	public void loadSnap(String source)
		throws IOException, FileNotFoundException
	{
		byte [] data = new byte[65536];
		int len;
		int n;

		FileInputStream fis = new FileInputStream(source);
		len = fis.read(data);
		fis.close();

		for (n = 0; n < len; n++)
		{
			if (data[n]   == (byte)0x8E && data[n+1] == (byte)0x49 &&
			    data[n+2] == (byte)0xC2 && data[n+3] == (byte)0xB6 &&
			    data[n+4] == (byte)0x49 && data[n+5] == (byte)0x83)
			{
//				System.err.println("Found Manic Miner at offset " + Integer.toHexString(n - 0x3A6E));
				System.arraycopy(data, n - 0x3A6E,
						dragonMem, 0, 0x8000);

				return;
			}
		}
		throw new IOException("Could not find Manic Miner room loader in " + source);
	}


	protected final int peek(int a)
	{
		int v1 = dragonMem[a+1];
		int v0 = dragonMem[a];
		return ((v0 << 8) & 0xFF00) | (v1 & 0xFF);
	}

	private final String hex(byte b)
	{
		String s = Integer.toHexString(b & 0xFF);
		while (s.length() < 2) s = "0" + s;
		return s;
	}
	
	
	private final String hex4(int b)
	{
		String s = Integer.toHexString(b & 0xFFFF);
		while (s.length() < 4) s = "0" + s;
		return s;
	}

	protected void dumpRoom(int a)
	{
		int g, mg;
		StringBuffer buf = new StringBuffer(512);
		
		buf.setLength(512);
		for (g = 0; g < 512; g++) buf.setCharAt(g,' ');

		System.out.println("\nRoom " + Integer.toString(a));	
		mg = dragonMem[a++];
		System.out.println(hex((byte)mg));
		for (g = 0; g < mg; g++)
		{
			System.out.print("Guardian: ");
			for (int n = 0; n < 11; n++)
			{
				System.out.print(hex(dragonMem[a++]));
				System.out.print(" ");
			}	
			System.out.println();
		}
		mg = dragonMem[a++];
		System.out.print("Vertical: " + hex((byte)mg) + " ");
		if (mg == 0) 
		{
			System.out.println("");
		}
		else
		{
			System.out.println(hex(dragonMem[a++]));
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];
				int offset = (y-1)*32 + (x-1);
	
				System.out.print(hex(y));
				System.out.print(" ");
				System.out.print(hex(x));
				System.out.print(" ");
				System.out.println(c);
				while (c != 0)
				{
					if (offset >= 0 && offset < 512)
					buf.setCharAt(offset, 'v');
					offset += 32;
					--c;
				}	
			}
		}
		mg = dragonMem[a++];
		System.out.print("Water: ");
		if (mg == 0)
		{
			System.out.println(hex((byte)mg));
		}
		else
		{
			System.out.print(hex((byte)mg) + " ");
			System.out.println(hex(dragonMem[a++]));
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];
				int offset = (y-1)*32 + (x-1);
	
				System.out.print(hex(y));
				System.out.print(" ");
				System.out.print(hex(x));
				System.out.print(" ");
				System.out.println(c);
				while (c != 0)
				{
					if (offset >= 0 && offset < 512)
					buf.setCharAt(offset, 'w');
					offset ++;
					--c;
				}	
			}
		}
		mg = dragonMem[a++];
		System.out.print("Crumbly: ");
		if (mg == 0)
		{
			System.out.println(hex((byte)mg));
		}
		else
		{
			System.out.print(hex((byte)mg) + " ");
			System.out.println(hex(dragonMem[a++]));
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];
				int offset = (y-1)*32 + (x-1);
	
				System.out.print(hex(y));
				System.out.print(" ");
				System.out.print(hex(x));
				System.out.print(" ");
				System.out.println(c);
				while (c != 0)
				{
					if (offset >= 0 && offset < 512)
					buf.setCharAt(offset, 'c');
					offset ++;
					--c;
				}	
			}
		}
		mg = dragonMem[a++];
		System.out.print("Conveyor: ");
		if (mg == 0)
		{
			System.out.println(hex((byte)mg));
		}
		else
		{
			System.out.print(hex((byte)mg) + " ");
			byte ctype = dragonMem[a++];
			System.out.println(hex(ctype));
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];
				int offset = (y-1)*32 + (x-1);
	
				System.out.print(hex(y));
				System.out.print(" ");
				System.out.print(hex(x));
				System.out.print(" ");
				System.out.println(c);
				while (c != 0)
				{
					if (offset >= 0 && offset < 512)
					buf.setCharAt(offset, (ctype == 0) ? '>' : '<');
					offset ++;
					--c;
				}	
			}
		}
		mg = dragonMem[a++];
		System.out.print("Earth V: ");
		System.out.println(hex((byte)mg));
		if (mg != 0)
		{
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];
				int offset = (y-1)*32 + (x-1);
	
				System.out.print(hex(y));
				System.out.print(" ");
				System.out.print(hex(x));
				System.out.print(" ");
				System.out.println(c);
				while (c != 0)
				{
					if (offset >= 0 && offset < 512)
					buf.setCharAt(offset, '#');
					offset += 32;
					--c;
				}	
			}
		}
		mg = dragonMem[a++];
		System.out.print("Earth H: ");
		System.out.println(hex((byte)mg));
		if (mg != 0)
		{
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];
				int offset = (y-1)*32 + (x-1);
	
				System.out.print(hex(y));
				System.out.print(" ");
				System.out.print(hex(x));
				System.out.print(" ");
				System.out.println(c);
				while (c != 0)
				{
					if (offset >= 0 && offset < 512)
					buf.setCharAt(offset, '#');
					offset ++;
					--c;
				}	
			}
		}
		mg = dragonMem[a++];
		System.out.print("Items: ");
		if (mg == 0)
		{
			System.out.println(hex((byte)mg));
		}
		else
		{
			System.out.print(hex((byte)mg) + " ");
			byte ctype = dragonMem[a++];
			System.out.println(hex(ctype));	
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				int offset = (y-1)*32 + (x-1);
	
				System.out.print(hex(y));
				System.out.print(" ");
				System.out.println(hex(x));
				if (offset >= 0 && offset < 512)
				buf.setCharAt(offset, '*');
			}
		}
		mg = dragonMem[a++];
		System.out.print("Fire: ");
		System.out.println(hex((byte)mg));
		if (mg != 0)
		{
			for (g = 0; g < mg; g++)
			{
				byte ctype = dragonMem[a++];
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				int offset = (y-1)*32 + (x-1);
	
				System.out.print(hex(ctype));
				System.out.print(" ");
				System.out.print(hex(y));
				System.out.print(" ");
				System.out.println(hex(x));
				if (offset >= 0 && offset < 512)
				buf.setCharAt(offset, '!');
			}
		}
		byte py = dragonMem[a++];
		byte px = dragonMem[a++];
		int ppos = (py-1)*32 + (px-1);
		buf.setCharAt(ppos, 'P');
		buf.setCharAt(ppos+1, 'P');
		buf.setCharAt(ppos+32, 'P');
		buf.setCharAt(ppos+33, 'P');
		py = dragonMem[a++];
		px = dragonMem[a++];
		ppos = (py-1)*32 + (px-1);
		buf.setCharAt(ppos, 'W');
		buf.setCharAt(ppos+1, 'W');
		buf.setCharAt(ppos+32, 'W');
		buf.setCharAt(ppos+33, 'W');
		
		py = dragonMem[a++];
		px = dragonMem[a++];
		System.out.println("Unknown 1: " + hex(py));
		System.out.println("Unknown 2: " + hex(px));
		int titleLen = dragonMem[a++];
		StringBuffer title = new StringBuffer();
		mg = (32 - titleLen) / 2;
		for (g = 0; g < 32; g++) 
		{
			if (g < mg || g >= (mg + titleLen)) title.append(' ');
			else title.append(charSet.charAt(dragonMem[a++]));
		}
		System.out.println("Title: \"" + title.toString() + "\"");
		py = dragonMem[a++];
		px = dragonMem[a++];
		System.out.println("Unknown 3: " + Integer.toHexString(px + 256 * (0xFF & py)));

		String s = buf.toString();
		for (g = 0; g < 16; g++)
		{
			System.out.println(s.substring(g*32,g*32+32));
		}
	}	
	
	public void dumpRooms()
	{
		for (int a = 0x49C2; a < 0x49ee; a += 2)
		{
			int address = peek(a);
			dumpRoom(address);
		}
		for (int b = 0; b < 0x49C2; b++)
		{
			byte v = dragonMem[b];
	
			if ((b & 0x3F) == 0) 
			{
				System.out.print(hex4(b) + ": ");
			}	
			if ((0xFF & v) < charSet.length())
			{
				System.out.print(charSet.charAt(v));	
			}
			else
			{
				System.out.print(".");
			}
			if ((b & 0x3F) == 0x3F) System.out.println();
		}
	}

	public void setVariant(char v)
	{
		target.setVariant(v);
	}

	public void load(String dragon, String spectrum)
		throws IOException, FileNotFoundException
	{
		loadSnap(dragon);
		target.load48(target.getVariant(), spectrum);
	}

	public void save(String filename) throws IOException
	{
		target.saveTape(filename);
	}

	public static void main(String args[])
	{
		boolean dump = false;
		char variant = 'w';

		try
		{
			if (args.length < 2)
			{
				System.err.println("Syntax: java DragonManic manic.dragonMem tapfile");
				return;
			}
			for (int n = 2; n < args.length; n++)
			{
				if (args[n].charAt(0) == '-' &&
				    args[n].charAt(1) == 'f')
				{
					variant = args[n].charAt(2);
				}	
				if (args[n].charAt(0) == '-' &&
				    args[n].charAt(1) == 'd')
				{
					dump = true;
				}
			}
			DragonManic dm = new DragonManic(args[0]);
			if (dump)
			{
				dm.dumpRooms();
			}
			dm.target.load48(variant, "jsw.sna");
			dm.execute();
			dm.save(args[1]);
		}
		catch(FileNotFoundException e)
		{
			System.err.println("File not found: " + e.getMessage());
		}
		catch(IOException e)
		{
			System.err.println("I/O exception: " + e.getMessage());
		}
	}

	protected void convertRoom(int a, int nroom)
	{
		int g, mg;
		byte [] cellData = new byte[9];

		cellData[0] = AIR;
		target.setCellPattern(nroom, Jsw64Game.AIR, 1, cellData);
		target.setWillyColour(nroom, 7);
		target.setKamikazeFlag(nroom, true);	
		target.setAir(nroom, 161);

		mg = dragonMem[a++];
		// Guardians.
		if (mg != 0)
		{
			byte g2[];
			byte guards[] = new byte[mg * 8 + 1];
			byte dg[]     = new byte[11];
			for (g = 0; g < mg; g++)
			{
				for (int n = 0; n < 11; n++)
				{
					dg[n] = dragonMem[a++];
				}
				if (dg[0] == 0) // Horizontal
				{
					if ((8 * g) >= guards.length) continue;
					boolean biDirectional = true;
// XXX Initial frame is probably (dg[5] - 1) & 0xE0
					int frame = ((dg[5] - 1) >> 5);
					guards[8*g  ] = 1;   // Horizontal
					guards[8*g+1] = 0x0E;// Colour
					guards[8*g+2] = (byte)(((dg[5] - 1)) & 0x1F);
					guards[8*g+3] = (byte)((dg[1] - 1) * 16);
					guards[8*g+4] = 0;
					guards[8*g+5] = (byte)convertSprites(dg, biDirectional);
					guards[8*g+6] = (byte)(dg[2] - 1);
					guards[8*g+7] = (byte)(dg[3] - 2);
					if (biDirectional) guards[8*g+1] |= 0xE0;
					else		   guards[8*g+1] |= 0x60;
					if (dg[7] == 2)	// Slow
					{
						guards[8*g+7] |= 0x20;
					}
					if (dg[4] == 0)	// Going right?
					{
						guards[8*g] |= 0x80;
					}
				}
				else if (dg[4] > 8) // Skylab
				{
// On the first skylab, limit ourselves to migrating only three guardians
					if (g == 0)
					{
						guards = new byte[25];
						guards[24] = -1;
					}
					if (g >= 3) continue;
					boolean biDirectional = false;
// XXX Initial frame is probably (dg[5] - 1) & 0xE0
					int frame = ((dg[5] - 1) >> 5);
					guards[8*g  ] = 0x08; // Vertical
					guards[8*g+1] = 0x0E;// Colour
					guards[8*g+2] = (byte)(((dg[1] - 1)) & 0x1F);
					guards[8*g+3] = 0;
					guards[8*g+4] = (byte)(dg[9] * 2);
					guards[8*g+5] = (byte)convertSprites(dg, biDirectional);
					guards[8*g+6] = 0;
					guards[8*g+7] = (byte)((dg[3] - 1)* 16);
				}
				else	// Vertical
				{
					if ((8 * g) >= guards.length) continue;
					boolean biDirectional = false;
// XXX Initial frame is probably (dg[5] - 1) & 0xE0
					int frame = ((dg[5] - 1) >> 5);
					guards[8*g  ] = 2;   // Vertical
					guards[8*g+1] = 0x6E;// Colour+animation
					guards[8*g+2] = (byte)(((dg[1] - 1)) & 0x1F);
					guards[8*g+3] = (byte)((dg[5] - 1) * 16);
					guards[8*g+4] = (byte)(dg[9] * 2);
					guards[8*g+5] = (byte)convertSprites(dg, biDirectional);
					guards[8*g+6] = (byte)((dg[2] - 1)* 16);
					guards[8*g+7] = (byte)((dg[3] - 1)* 16);
					if (dg[4] == 0)	// Going up?
					{
						guards[8*g+4] = (byte)(-guards[8*g+4]);
					}
				}
			}
			guards[guards.length - 1] = -1;
			switch(nroom)
			{
				case 4: // Eugene's Lair. Add a trigger
					// that drives Eugene wild.
					g2 = new byte[guards.length + eugene.length];
					System.arraycopy(guards, 0, g2, eugene.length,
							guards.length);
					System.arraycopy(eugene, 0, g2, 0, eugene.length);
					// Keep same sprite page.
					g2[5] = g2[13];
					guards = g2;
					break;
				case 7:  // Kong rooms.
				case 11:
					g2 = new byte[guards.length + kong.length];
					System.arraycopy(kong, 0, g2, 0, kong.length);
					System.arraycopy(guards, 0, g2, kong.length,
							guards.length);
					g2[29] = guards[5]; // Keep sprite page
					g2[30] = guards[6]; // and limit
					g2[31] |= (guards[7] & 0xE0);
					guards = g2;
					break;
				case 10: // Mutant telephones
					guards[6] += 2;	// Limit on guard 0
					break;		// to avoid collision
				case 14: // The Bank
					guards[7]--;	// As above
					break;
			}
			target.putGuardians(nroom, guards);
		}
		// Vertical water 2
		mg = dragonMem[a++];
		if (mg != 0) 
		{
			byte cell = dragonMem[a++];
			int src = peek(0x4A04 + 2*cell);
			cellData[0] = WATER2;
			System.arraycopy(dragonMem, src, cellData, 1, 8);

			target.setCellPattern(nroom, Jsw64Game.WATER, 2,
					cellData);

			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];

				while (c != 0)
				{
					target.setCell(nroom, x - 1, y - 1, 
						Jsw64Game.WATER, 2);
					++y;
					--c;
				}
			}
		}
		// Horizontal water 1
		mg = dragonMem[a++];
		if (mg != 0)
		{
			byte cell = dragonMem[a++];
			int src = 0x4EB7 + 8 * cell;
			cellData[0] = WATER1;
			System.arraycopy(dragonMem, src, cellData, 1, 8);

			target.setCellPattern(nroom, Jsw64Game.WATER, 1,
					cellData);

			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];

				while (c != 0)
				{
					target.setCell(nroom, x - 1, y - 1, 
						Jsw64Game.WATER, 1);
					++x;
					--c;
				}
			}
		}
		mg = dragonMem[a++];
		if (mg != 0)
		{
			byte cell = dragonMem[a++];
			int src = 0x4EB7 + 8 * cell;
			cellData[0] = CRUMBLY;
			System.arraycopy(dragonMem, src, cellData, 1, 8);
// Manual fixup for the fast-crumble bug
			if (cell == 6)
			{
				cellData[8] = (byte)0x80;
				cellData[7] = 1;
				cellData[5] = (byte)0x80;
				cellData[3] = 1;
			}

			target.setCellPattern(nroom, Jsw64Game.CRUMBLY, 1,
					cellData);

			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];

				while (c != 0)
				{
					target.setCell(nroom, x - 1, y - 1, 
						Jsw64Game.CRUMBLY, 1);
					++x;
					--c;
				}
			}
		}
		mg = dragonMem[a++];
		if (mg != 0)
		{
			byte ctype = dragonMem[a++];
			int cdir;
			int src = 0x4E97;
			cellData[0] = CONVEYOR;
			System.arraycopy(dragonMem, src, cellData, 1, 8);

			if (ctype == 0) cdir = Jsw64Game.RCONVEYOR;
			else		cdir = Jsw64Game.LCONVEYOR;
			target.setCellPattern(nroom, cdir, 1, cellData);
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];

				target.setConveyor(nroom, x-1, y-1, c, (ctype != 0));
				while (c != 0)
				{
					target.setCell(nroom, x-1, y-1, cdir, 1);
					++x;
					--c;
				}
			}
		}
		mg = dragonMem[a++];
		if (mg != 0)
		{
			int src = 0x4F0F;
			cellData[0] = EARTH;
			System.arraycopy(dragonMem, src, cellData, 1, 8);

			target.setCellPattern(nroom, Jsw64Game.EARTH, 1, cellData);
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];
	
				while (c != 0)
				{
					target.setCell(nroom, x-1, y-1, 
						Jsw64Game.EARTH, 1);
					++y;
					--c;
				}
			}
		}
		mg = dragonMem[a++];
		if (mg != 0)
		{
			int src = 0x4F0F;
			cellData[0] = EARTH;
			System.arraycopy(dragonMem, src, cellData, 1, 8);

			target.setCellPattern(nroom, Jsw64Game.EARTH, 1, cellData);
			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];
				byte c = dragonMem[a++];
	
				while (c != 0)
				{
					target.setCell(nroom, x-1, y-1, 
						Jsw64Game.EARTH, 1);
					++x;
					--c;
				}
			}
		}
		mg = dragonMem[a++];
		if (mg != 0)
		{
			byte itype = dragonMem[a++];
			int src = peek(0x49EE + 2 * itype);
			byte ibits[] = new byte[8];
			System.arraycopy(dragonMem, src, ibits, 0, 8);
			target.setItemBitmap(nroom, ibits);

			for (g = 0; g < mg; g++)
			{
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];

				target.addItem(nroom, x-1, y-1);	
				// Ensure the item isn't on top of something
				target.setCell(nroom, x-1, y-1, Jsw64Game.AIR, 1);
			}
		}
		mg = dragonMem[a++];
		if (mg != 0)
		{
			int ct1, ct2;
			ct1 = -1;
			ct2 = -1;
			for (g = 0; g < mg; g++)
			{
				byte ctype = dragonMem[a++];
				byte y = dragonMem[a++];
				byte x = dragonMem[a++];

				int src = 0x4F17 + 8 * ctype;
				int idx = -1;
				if (ct1 == -1 || ct1 == src) 
				{
					ct1 = src; idx = 1;
				}
				else if (ct2 == -1 || ct2 == src)
				{
					ct2 = src; idx = 2;
				}
				if (idx == -1) continue;
				cellData[0] = (byte)((idx == 1) ? FIRE1 : FIRE2);
				System.arraycopy(dragonMem, src, cellData, 1, 8);
				target.setCellPattern(nroom, Jsw64Game.FIRE, idx, 
						cellData);
				target.setCell(nroom, x-1, y-1, Jsw64Game.FIRE, idx);
			}
		}
		byte py = dragonMem[a++];
		byte px = dragonMem[a++];
		target.setPortalPos(nroom, px-1, py-1);
		target.setPortalSprite(nroom, 0xA680);
		target.setPortalAttr(nroom, (byte)PORTAL);
		py = dragonMem[a++];
		px = dragonMem[a++];
		if (nroom == 0)
		{
			target.setStartRoom(nroom, px-1, py-1);
			target.setPortalDest(21, nroom, px-1, py-1);
			target.setExit(nroom, Jsw64Game.JSW64_EXIT_L, 21);
			target.setExit(nroom, Jsw64Game.JSW64_EXIT_U, nroom);
			target.setExit(nroom, Jsw64Game.JSW64_EXIT_D, nroom);
			target.setExit(nroom, Jsw64Game.JSW64_EXIT_R, (nroom + 1) % 22);
		}
		else
		{
			target.setPortalDest(nroom - 1, nroom, px-1, py-1);
			target.setExit(nroom, Jsw64Game.JSW64_EXIT_L, nroom - 1);
			target.setExit(nroom, Jsw64Game.JSW64_EXIT_U, nroom);
			target.setExit(nroom, Jsw64Game.JSW64_EXIT_D, nroom);
			target.setExit(nroom, Jsw64Game.JSW64_EXIT_R, (nroom + 1) % 22);
		}
		++a;	// Unknown
		++a;	// Unknown
		int titleLen = dragonMem[a++];
		String charSet = " ABCDEFGHIKLMNOPRSTUVWXYJ";
		StringBuffer title = new StringBuffer();
		mg = (32 - titleLen) / 2;
		for (g = 0; g < 32; g++) 
		{
			if (g < mg || g >= (mg + titleLen)) title.append(' ');
			else title.append(charSet.charAt(dragonMem[a++]));
		}
		target.setRoomTitle(nroom, title.toString());
		switch(nroom)
		{
			case 7: case 11:	// Kong rooms
				target.setInitPatch(nroom, 0x9800);
				target.setLoopPatch(nroom, 0x9803);
				break;
			case 9: // The Endorian Forest
				// Avoid a guardian collision
				target.setCell(nroom, 23, 5, Jsw64Game.AIR, 1);					break;
			case 21:		// Real final barrier
				target.setPortalPatch(nroom, 0x9806);
				break;
		}
	}	

	private int convertSprites(byte[] guardian, boolean biDirectional)
	{
		int map, spriteSrc, ss0, frame;
		int x, y;

 		if (guardian[0] == 0)     spriteSrc = (guardian[6] * 256) + 24403;	
		else if (guardian[4] > 8) spriteSrc = 29907;
		else	                  spriteSrc = (guardian[6] * 128) + 28755;
		ss0 = spriteSrc;
		for (map = 0; map < spriteMap.size(); map++)
		{
			Integer i = (Integer)spriteMap.elementAt(map);
			if (i.intValue() == spriteSrc)
			{
				return spritePage[map];
			}
		}
		if (spriteSrc >= dragonMem.length)
		{
			return 0xA6;	// XXX Something wrong here
		}
		byte page[] = new byte[256];

		for (frame = 0; frame < 8; frame++)
		{
			int f1 = frame;
			if (biDirectional) f1 = frame ^ 4;
			
			for (x = 1; x >= 0; x--)
			{
				for (y = 0; y < 16; y++)
				{
					byte v = dragonMem[spriteSrc++];
					page[32*f1+2*y+x] = v;
				}
			}
		}
		if (guardian[0] == 0) rotateFrames(page);
		target.storeSpritePage(spritePage[map], page);
		spriteMap.addElement(new Integer(ss0));
		return spritePage[map];
	}

	void rotateFrames(byte[] page)
	{
		int frame, x, y; 

		for (frame = 0; frame < 8; frame++)
		{
			int rr = 0, rl = 0;
			int f1 = frame ^ 4;
			switch(frame & 3)
			{
				case 0: break;
				case 1: rr = 2; break;
				case 2: rr = 4; break;
				case 3: rr = 6; break;
			}
			if (rr != 0 || rl != 0) for (y = 0; y < 16; y++)
			{
				byte b1 = page[32*f1+2*y]; 
				byte b2 = page[32*f1+2*y+1]; 

				if (rr != 0) for (x = 0; x < rr; x++)
				{
					if ((b1 & 1) != 0)
					{
						b2 = (byte)((b2 >> 1) | 0x80);
					}
					else
					{
						b2 = (byte)((b2 >> 1) & 0x7F);
					}
					b1 = (byte)((b1 >> 1) & 0x7F);
				}
				if (rl != 0) for (x = 0; x < rl; x++)
				{
					if ((b2 & 0x80) != 0)
					{
						b1 = (byte)((b1 << 1) | 1);
					}
					else
					{
						b1 = (byte)((b1 << 1) & 0xFE);
					}
					b2 = (byte)((b2 << 1) & 0xFE);
				}
				page[32*f1 + 2*y] = b1;
				page[32*f1 + 2*y+1] = b2;
			}
		}
	}

	InputStream openPatch(String s) throws FileNotFoundException
	{
		InputStream is = getClass().getResourceAsStream(s);
		if (is == null) throw new FileNotFoundException(s);
		return is;
	}

	public void execute() throws IOException, FileNotFoundException
	{
		InputStream is;
		spriteMap = new Vector();
		final byte[] elementMap = 
		{
			0, 1, 2, 3, 4, 5, 6, 7,
			8, 2, 0, 1, 3, 2, 1, 1
		};
		target.setElementMap(elementMap);
		target.clearItems();
		target.loadPatch(is = openPatch("dmm64.bin"), 0x9800);
		is.close();
		// Migrate portal sprite to A680, switch to A6A0.
		byte[] page = target.getSpritePage(0xA6);
		for (int y = 0; y < 16; y++) 
		{
			page[0xA0 + 2*y] = dragonMem[0x4FCF + y];
			page[0xA0 + 2*y+1] = 0;
			for (int x = 0; x < 2; x++)
			{
				page[0x80 + 2*y+x] = dragonMem[0x4FFF + 16*(1-x)+y];
			}
		}
		target.storeSpritePage(0xA6, page);
		// Migrate plinth sprite to 9C60, boot to 9C40.
		page = target.getSpritePage(0x9C);
		for (int y = 0; y < 16; y++) 
			for (int x = 0; x < 2; x++)
			{
				page[0x60 + 2*y+x] = dragonMem[0x501F + 16*(1-x)+y];
				page[0x40 + 2*y+x] = dragonMem[0x503F + 16*(1-x)+y];
			}
		target.storeSpritePage(0x9C, page);
		// Copy Kong to 9B80.
		page = target.getSpritePage(0x9B);
		for (int s = 0; s < 4; s++)
			for (int y = 0; y < 16; y++) 
				for (int x = 0; x < 2; x++)
			{
				page[0x80+32*s+2*y+x] = dragonMem[0x7453+32*s+16*(1-x)+y];
			}
		target.storeSpritePage(0x9B, page);
		// Migrate Willy's sprite
		page = target.getSpritePage(0x9D);
		for (int s = 0; s < 8; s++)
		{
			for (int y = 0; y < 16; y++)
				for (int x = 0; x < 2; x++)
				{
					page[32 * s + 2*y + x] =
						dragonMem[0x4D74+mapFrame[s]*32+16*(1-x)+y];
				}
		}
		rotateFrames(page);
		target.storeSpritePage(0x9D, page);
		int nr = 0;
		for (nr = 0; nr < 22; nr++)
		{
			target.initRoom(nr);	
		}
		nr = 0;
		for (int a = 0x49C2; a < 0x49ee; a += 2)
		{
			int address = peek(a);
			convertRoom(address, nr++);
		}
		target.setMemoryMap(0x98, (byte)0x80);
		target.setMemoryMap(0x9E, (byte)0x80);
		target.setMemoryMap(0x9F, (byte)0x80);
		// Patch various bits of the game engine. 
		// Super-mono patch (mono screen but flashing portals)
		target.poke(0x8a2C, (byte)0xCD);
		target.poke(0x8a2D, (byte)0x18);
		target.poke(0x8a2E, (byte)0x98);
		target.poke(0x8a2F, (byte)0x00);
		target.poke(0x8a30, (byte)0x00);
		// Conveyors. Firstly, drop the check
		// for Maria and the toilet as these are JSW-specific.
		target.poke(0x89E6, (byte)0x00);
		target.poke(0x89E7, (byte)0x00);
		target.poke(0x89E8, (byte)0x00);
		target.poke(0x89E9, (byte)0x00);
		target.poke(0x89EA, (byte)0x00);
		target.poke(0x89EB, (byte)0x00);
		// Then set the conveyor routine to be our substitute.
		target.poke(0x89F0, (byte)0x1E);
		target.poke(0x89F1, (byte)0x98);
		// Item flashing
		target.poke(0x944E, (byte)0xCD);
		target.poke(0x944F, (byte)0x1B);
		target.poke(0x9450, (byte)0x98);
		// Disallow falling off the bottom; The last room in the
		// Dragon version relies on Willy not falling off.
		target.poke(0x8E47, (byte)0xD4);
		target.poke(0x8E48, (byte)0x8E);
		// Wrap at left and right
		target.poke(0x948D, (byte)0x00);	// Nop out the bit that 
		target.poke(0x948E, (byte)0x00);	// changes the room number.
		target.poke(0x948F, (byte)0x00);	// (left exit)
		target.poke(0x949A, (byte)0xC9);
		target.poke(0x94A1, (byte)0x00);	// Nop out the bit that
		target.poke(0x94A2, (byte)0x00);	// changes the room number.
		target.poke(0x94A3, (byte)0x00);	// (right exit)
		target.poke(0x94AC, (byte)0xC9);

		// Font
		target.loadPatch(is = openPatch("dffix.bin"), 0xC100);
		is.close();
		for (int n = 0; n < charSet.length(); n++)
		{
			int ct = (int)(charSet.charAt(n)) * 8 + 0xC000;
			int source = 0x4BB9 + 8 * n;
			for (int m = 0; m < 8; m++)
			{
				target.poke(ct + m, dragonMem[source + m]);
			}
		}
		// Add the 'weak trigger' guardian
		target.poke(0xEC98 + 26, (byte)9);
		target.poke(0xEC98 + 27, (byte)0x98);
		// Make messages upper-case
		target.poke(0x8454, "      PRESS ENTER TO START      ");
		target.poke(0x8554, "ITEMS COLLECTED 000 TIME 00:00 MGAMEOVER");
		target.poke(0x8584, "A 7:00A");
		// Move the magic words to room 0.
		target.poke(0x8BA3, (byte)0);
		target.poke(0xC40D, (byte)0);
		target.poke(0x85E7, typewriter);
		target.poke(0x87E1, (byte)2);	// Lives
		// Make bottom third of screen mono
		for (int n = 0x97E0; n < 0x9800; n++) target.poke(n, (byte)0x47);
		for (int n = 0x9A00; n < 0x9A20; n++) target.poke(n, (byte)0x78);	
		for (int n = 0x9A20; n < 0x9B00; n++) target.poke(n, (byte)0x47);	
		// Make bottom third of title screen mono.
		target.poke(0x8832, (byte)0x47);
		// Get rid of colours during pause.
		target.poke(0x8AE6, (byte)0);
		target.poke(0x8AE7, (byte)0);
		target.poke(0x8AE8, (byte)0);
		// Make the plinth white on the game-over screen.
		target.poke(0x8CB2, (byte)7);
		target.setFinalBarrier(0x13, true, Jsw64Game.FINAL_BARRIER_BUF2);
		makeTitle();	
		target.loadTitleTune(is = openPatch("danube.tun"));
		is.close();
	}

	static final byte [] title_hdr = {
		0x50, 0x4C, 0x55, 0x53, 0x33, 0x44, 0x4F, 0x53,
		0x1A, 0x01, (byte)0x80, 0x1B, 0x00, 0x00, 0x00, 0x03,
		0x00, 0x1B, 0x00, (byte)0xC0, 0x00, (byte)0x80
	};

	void makeTitle() throws IOException, FileNotFoundException
	{
		byte title[]  = new byte[7040];
		byte title7[] = new byte[2048+256];
		System.arraycopy(title_hdr, 0, title, 0, title_hdr.length);
		title[0x7F] = 0x71;
		int x, y;
		for (y = 0; y < 80; y++)
			for (x = 0; x < 32; x++)
			{
				int src = 0x1E00 + 32 * y + x;
				int dest = 0x80 +  32 * ((y%64)/8) + 
					        + 256 * (y%8)+x 
						+ 2048 * (y/64);
				title[dest] = dragonMem[src];
			}
		for (y = 4096; y < 5632; y++) title[y + 0x80] = 0x47;

		for (y = 0x47D3; y <= 0x4906; y++)
		{
			title[0x1680 + y - 0x47D3] = 
				(byte)(charSet.charAt(dragonMem[y]));
		}

		System.arraycopy(title, 128, title7, 0, 2048);
		for (y = 0; y < 256; y++) title7[y + 2048] = 0x47;
		for (y = 0; y < 19; y++) title7[y + 2048 + 224] = EARTH;
		for (y = 19; y < 31; y++)
		{
			title7[y + 2048 + 224] = AIR;
			title7[y + 2048 + 192] = AIR;
			title7[y + 2048 + 160] = AIR;
		}
		title7[2048 + 224 + 21] = EARTH;
		title7[2048 + 224 + 31] = EARTH;
		title7[2048 + 192 + 18] = EARTH;
		title7[2048 + 192 + 21] = EARTH;
		title7[2048 + 192 + 31] = EARTH;
		title7[2048 + 160 + 18] = EARTH;
		title7[2048 + 160 + 21] = EARTH;
		title7[2048 + 160 + 31] = EARTH;
		for (y = 18; y < 32; y++) title7[y + 2048 + 128] = EARTH;

		target.setFinalBarrierGraphic(false, title7);
		target.setFinalBarrierGraphic(true, title7);
		target.setTitleScreen(title, 128);

//		FileOutputStream fos = new FileOutputStream("title.jw7");
//		fos.write(title);
//		fos.close();
	}



}
