/*
    HEXPAT: Patch an object file with an Intel Hex overlay
    Copyright (C) 2004, 2006 John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    This program is based on the public domain LOAD.C from the Simtel-20
    UNIX-C collection. Its original author is unknown.  

    LOAD.C
	Convert a .HEX format file to a .COM format file.
	Contributor: unknown

   By this stage, my code outnumbers the original code by a factor of about 3.

*/

package info.seasip.jsw64;

import java.io.*;

public class HexPatcher
{
	byte [] memory;

	HexPatcher(byte [] mem)
	{
		memory = mem;
	}

	byte hex_checksum, tap_checksum;
	int hexline = 0;

	int gethex(InputStream is) throws IOException
	{
		int    c;
		int    x;

		c = is.read();
		if ('0' <= c && c <= '9') 	x = c - '0';
		else if ('A' <= c && c <= 'F')	x = c - 'A' + 10;
		else 
		{
			throw(new IOException("Funny hex letter 0x" + Integer.toHexString(c)));
		}
		x <<= 4;
		c = is.read();
		if ('0' <= c && c <= '9') 	x |= c - '0';
		else if ('A' <= c && c <= 'F')	x |= c - 'A' + 10;
		else 
		{
			throw(new IOException("Funny hex letter 0x" + Integer.toHexString(c)));
		}
		hex_checksum += x;
		return x;
	}



	void patch(InputStream is) throws IOException
	{
		hexline = 0;
		int c, n, i, ah, al, naddr;	
		int type;

		do {
/* Skip until the first hex record starts. Some .HEX files have a line
 * explaining what they are. */
			do {
		 		c = is.read();
		    		if (c == -1)
				{
					throw new IOException("Premature EOF on HEX file");
	    			}
			} while (c != ':');
			++hexline;
			hex_checksum = 0;
			n  = gethex(is);		/* bytes / line */
			ah = gethex(is);
			al = gethex(is);
	
			switch (type = gethex(is)) 
			{
			    case 0:
				if (n == 0) /* MAC uses a line with no bytes as EOF */
				{
					type = 1;
					break;
				}
				naddr = (ah << 8) | al;
				for (i = 0; i < n; i++)
				{
					memory[naddr] = (byte)gethex(is);
					++naddr;
				}
				break;
	
				case 1: break; 
				default:
					throw (new IOException("Line " + Integer.toString(hexline) + ": Funny record type " + Integer.toString(type)));
			}
			gethex(is);	/* Checksum byte */
	
			if (hex_checksum != 0) 
			{
			    throw (new IOException("Line " + Integer.toString(hexline) + ": Checksum error"));
			}
		} while (type != 1);
	}
}
