
package info.seasip.jsw64;

import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;

class ImagePanel extends Canvas 
{
	Image picture;
	int iw = -1, ih = -1;

	ImagePanel(Image i)
	{
		picture = i;
		prepareImage(picture, this);
		iw = picture.getWidth(this);
		ih = picture.getHeight(this);
		if (iw > 0 && ih > 0) setSize(iw + 2, ih + 2);
	}

	public boolean imageUpdate(Image image, int infoflags,
			int x, int y, int w, int h)
	{
		if ((infoflags & ImageObserver.WIDTH) != 0 && iw < 0) 
			iw = picture.getWidth(this);
		if ((infoflags & ImageObserver.HEIGHT) != 0 && ih < 0)
			ih = picture.getWidth(this);
		return super.imageUpdate(image, infoflags, x, y, w, h);
	}


	public Dimension getMinimumSize()
	{
		return new Dimension(iw + 2, ih + 2);
	}	
	
	public Dimension getPreferredSize()
	{
		return new Dimension(iw + 2, ih + 2);
	}

	public void paint(Graphics g)
	{
		super.paint(g);
		if (iw < 0 || ih < 0) return;

		Dimension d = getSize();
		int xoff = (d.width  - iw) / 2;
		int yoff = (d.height - ih) / 2;
		g.drawImage(picture, xoff, yoff, getBackground(), this);
	}
}

class AlertDialog extends Dialog implements ActionListener
{
	private String caption;
	private String action;
	private Image picture;

	AlertDialog(Frame f, String s, String a)
	{
		super(f, "Install JSW64:Dragon");
		caption = s;
		action = a;
		setModal(true);
		
		picture = loadPicture(a + ".gif");
	}

	private Image loadPicture(String name)
	{
		try
		{
			InputStream is = getClass().getResourceAsStream(name);
			if (is == null) return null;
			byte[] temp = new byte[1024];
			ByteArrayOutputStream os = new ByteArrayOutputStream();

			for (int l = is.read(temp); l > 0; l = is.read(temp))
			{
				os.write(temp, 0, l);
			}
			is.close();
			os.close();
			return Toolkit.getDefaultToolkit().createImage(os.toByteArray());
		}
		catch(Exception e) { return null; }
	}

    
	public void showModal()
	{
		Panel p = new Panel();
		setLayout(new BorderLayout());
		p.setLayout(new FlowLayout());
		if (picture != null)
		{
			add("West", new ImagePanel(picture));
		}
		add("Center", new Label(caption));
		add("South", p);
		Button b = new Button(action);
		p.add(b);
		b.addActionListener(this);
		addWindowListener(new WindowAdapter()
		{
			public void windowClosing(WindowEvent e)
			{
				dispose();
			}
		});
		pack();
		super.show();
	}

	public void actionPerformed(ActionEvent e)
	{
		dispose();
	}
}

class Segment implements ActionListener
{
	private TextField text;
	private Label label;
	private Button button;
	private Panel panel1;
	private Panel panel2;
	private Frame frame;
	private int loadSave;

	Segment(Frame f, String s, String txt, int ls)
	{
		frame = f;
		text = new TextField(txt);
		label = new Label(s, Label.RIGHT);
		button = new Button("Browse");
		loadSave = ls;
		panel1 = new Panel();
		panel2 = new Panel();
		panel1.setLayout(new BorderLayout());
		panel2.setLayout(new FlowLayout(FlowLayout.LEFT));
	}

	void add(Container c)
	{
		c.add(label);
		panel2.add(button);
		panel1.add("Center", text);
		panel1.add("East", panel2);
		c.add(panel1);
		button.addActionListener(this);
	}

	public void actionPerformed(ActionEvent e)
	{
		FileDialog fd = new FileDialog(frame, label.getText(), loadSave);
		fd.setFile(text.getText());
		fd.setModal(true);
		fd.show();

		String s = fd.getFile();
		String d = fd.getDirectory();

		if (d == null || s == null) return;

		File f = new File(d, s);
		text.setText(f.getAbsolutePath());
	}

	String getText()
	{
		return text.getText();
	}

}

public class Install extends Frame implements ActionListener
{
	Segment jsw48;
	Segment mmdragon;
	Segment jswdragon;
	Segment mmresult;
	Segment jswresult;
	Button go, cancel;

	Install()
	{
		super("JSW64:Dragon Installer");
		addControls();
	}

	private void addControls()
	{
	/* XXX Use GridBagLayout here? */
		setBackground(SystemColor.control);	
		go = new Button("Install");
		cancel = new Button("Close");
		go.addActionListener(this);
		cancel.addActionListener(this);
		setLayout(new BorderLayout());
		Panel p = new Panel();
		Panel q = new Panel();
		p.setLayout(new GridLayout(5,2));
		q.setLayout(new FlowLayout());
		jsw48 = new Segment(this, "Path to .SNA snapshot of JSW48", "jsw.sna", FileDialog.LOAD);
		jswdragon = new Segment(this, "Path to Xroar snapshot of Dragon JSW", "jsw.dsna", FileDialog.LOAD);
		mmdragon = new Segment(this, "Path to Xroar snapshot of Dragon Manic Miner", "manic.dsna", FileDialog.LOAD);
		jswresult = new Segment(this, "Create JSW tapefile as", "jswdragon.tap", FileDialog.SAVE);
		mmresult = new Segment(this, "Create Manic Miner tapefile as", "mmdragon.tap", FileDialog.SAVE);
		jsw48.add(p);
		jswdragon.add(p);
		mmdragon.add(p);
		jswresult.add(p);
		mmresult.add(p);
		q.add(go);
		q.add(cancel);
		add("Center", p);
		add("South", q);
		pack();
		addWindowListener(new WindowAdapter()
		{
			public void windowClosing(WindowEvent e)
			{
				dispose();
				System.exit(0);
			}
		});
	}
	
	
	public void actionPerformed(ActionEvent e)
	{
		String s = e.getActionCommand();
		if (s == "Install")
		{
			install();
		}
		else if (s == "Close")
		{
			dispose();
			System.exit(0);
		}
	}

	void alert(String message, String action)
	{
		AlertDialog ad = new AlertDialog(this, message, action);

		ad.showModal();
	}

	void install()
	{
		try
		{
			DragonJSW dj = new DragonJSW();

			dj.setVariant('w');
			dj.load(jswdragon.getText(), jsw48.getText());
			dj.execute();
			dj.save(jswresult.getText());
		}
		catch(FileNotFoundException ex)
		{
			alert("Could not create JSW. File not found: "
					+ex.getMessage(), "Cancel");
			return;
		}
		catch(IOException ex)
		{
			alert("Could not create JSW. I/O error: "
					+ex.getMessage(), "Cancel");
			return;
		}
		try
		{
			DragonManic dm = new DragonManic();

			dm.setVariant('w');
			dm.load(mmdragon.getText(), jsw48.getText());
			dm.execute();
			dm.save(mmresult.getText());
		}
		catch(FileNotFoundException ex)
		{
			alert("Could not create Manic Miner. File not found: "
					+ex.getMessage(), "Cancel");
			return;
		}
		catch(IOException ex)
		{
			alert("Could not create Manic Miner. I/O error: "
					+ex.getMessage(), "Cancel");
			return;
		}
		alert("Files were created successfully", "OK");

	}

	public static void main(String args[])
	{
		Install i = new Install();

		i.show();
	}
}
