
package info.seasip.jsw64;

import java.io.*;

public class Plus3dos
{
	/** Is this a valid +3DOS file header? 
	 * @param header The candidate header.
	 * @return true if this is a valid +3DOS header, else false. */
	public static boolean isPlus3dos(byte[] header)
	{
		byte sum = 0;
		int n;
		final String p3dos = "PLUS3DOS\032";

		if (header.length < 128) return false;	
		for (n = 0; n < 127; n++) sum += header[n];
		if (sum != header[127])
		{
			return false;
		}
		for (n = 0; n < p3dos.length(); n++) 
			if (header[n] != p3dos.charAt(n)) return false;
		return true;
	}	

	/** Load a file into an array, skipping any +3DOS header.
	 * @param is The stream to load.
	 * @param dest The destination array */
	public static void loadStream(InputStream is, byte [] dest)
		throws IOException
	{
		loadStream(is, dest, 0, dest.length);
	}

	/** Load a file into an array, skipping any +3DOS header.
	 * @param is The stream to load.
	 * @param dest The destination array
	 * @param base Offset in the array to load at
	 * @param len Number of bytes to load */
	public static void loadStream(InputStream is, byte [] dest, 
				int base, int len) throws IOException
	{
		int bytes;
		byte header[] = new byte[128];

		bytes = is.read(header);
		if (bytes < 128)
		{
			if (len < bytes) bytes = len;
			System.arraycopy(header, 0, dest, base, bytes);
			is.close();
			return;
		}
		if (!isPlus3dos(header))
		{
			if (len < bytes) bytes = len;
			System.arraycopy(header, 0, dest, base, bytes);
			len -= bytes;
			base += bytes;
		}
		if (len > 0) is.read(dest, base, len);
		is.close();
		
	}
}
